package generator

import "fmt"
import "git.tebibyte.media/sashakoshka/fspl/llvm"
import "git.tebibyte.media/sashakoshka/fspl/entity"

func (this *generator) valueToLocation (value llvm.Value) llvm.Value {
	pointer := this.blockManager.newAllocaFront(value.Type())
	this.blockManager.NewStore(value, pointer)
	return pointer
}

func (this *generator) locationToValue (pointer llvm.Value, ty entity.Type) (llvm.Value, error) {
	irType, err := this.generateType(ty)
	if err != nil { return nil, err }
	return this.blockManager.NewLoad(irType, pointer), nil
}

// generateExpression generates an expression and returns a register
// representing the result.
func (this *generator) generateExpression (expression entity.Expression) (llvm.Value, error) {
	// we get an address from these, so we need to load the value
	switch expression := expression.(type) {
	case *entity.Dereference,
		*entity.MemberAccess,
		*entity.Slice,
		*entity.Subscript,
		*entity.LiteralArray,
		*entity.LiteralString,
		*entity.LiteralStruct:
		
		pointer, err := this.generateExpressionLoc(expression)
		if err != nil { return nil, err }
		return this.locationToValue(pointer, expression.Type())

	// we get a value from these, so we can return them as-is
	case *entity.Variable:
		return this.generateVariable(expression)
	case *entity.Declaration:
		return this.generateDeclaration(expression)
	case *entity.Call:
		return this.generateCall(expression)
	case *entity.MethodCall:
		return this.generateMethodCall(expression)
	case *entity.Reference:
		return this.generateReference(expression)
	case *entity.Length:
		return this.generateLength(expression)
	case *entity.ValueCast:
		return this.generateValueCast(expression)
	case *entity.BitCast:
		return this.generateBitCast(expression)
	case *entity.Operation:
		return this.generateOperation(expression)
	case *entity.Block:
		return this.generateBlock(expression)
	case *entity.IfElse:
		return this.generateIfElse(expression, false)
	case *entity.Loop:
		return this.generateLoop(expression, false)
	case *entity.Break:
		return this.generateBreak(expression)
	case *entity.Return:
		return this.generateReturn(expression)

	case *entity.LiteralInt:
		return this.generateLiteralInt(expression)
	case *entity.LiteralFloat:
		return this.generateLiteralFloat(expression)
	case *entity.LiteralBoolean:
		return this.generateLiteralBoolean(expression)
	default:
		panic(fmt.Sprintf (
			"BUG: generator doesnt know about expression %v, ty: %T",
			expression, expression))
	}
}

// generateExpressionLoc generates an expression and returns a register
// representing the address of the result. This function avoids adding alloca
// instructions whenever possible, trying to refer to the original data.
func (this *generator) generateExpressionLoc (expression entity.Expression) (llvm.Value, error) {
	switch expression := expression.(type) {

	// we get a value from these, so we need to store the value and then
	// return an address to that
	case *entity.Call,
		*entity.MethodCall,
		*entity.Reference,
		*entity.Length,
		*entity.ValueCast,
		*entity.BitCast,
		*entity.Operation,
		*entity.LiteralInt,
		*entity.LiteralFloat,
		*entity.LiteralBoolean:

		value, err := this.generateExpression(expression)
		if err != nil { return nil, err }
		return this.valueToLocation(value), nil

	// we get an address from these, so we can return them as-is
	case *entity.Variable:
		return this.generateVariableLoc(expression)
	case *entity.Declaration:
		return this.generateDeclarationLoc(expression)
	case *entity.Slice:
		return this.generateSliceLoc(expression)
	case *entity.Subscript:
		return this.generateSubscriptLoc(expression)
	case *entity.Dereference:
		return this.generateDereferenceLoc(expression)
	case *entity.MemberAccess:
		return this.generateMemberAccessLoc(expression)
	case *entity.Block:
		return this.generateBlockLoc(expression)
	case *entity.IfElse:
		return this.generateIfElse(expression, true)
	case *entity.Loop:
		return this.generateLoop(expression, true)
	case *entity.Break:
		return this.generateBreak(expression)
	case *entity.Return:
		return this.generateReturn(expression)
	case *entity.LiteralArray:
		return this.generateLiteralArrayLoc(expression)
	case *entity.LiteralString:
		return this.generateLiteralStringLoc(expression)
	case *entity.LiteralStruct:
		return this.generateLiteralStructLoc(expression)
	default:
		panic(fmt.Sprintf (
			"BUG: generator doesnt know about expression %v, ty: %T",
			expression, expression))
	}
}

func (this *generator) generateStatement (statement entity.Statement) (llvm.Value, error) {
	switch statement := statement.(type) {
	case *entity.Assignment:
		return this.generateAssignment(statement)
	case entity.Expression:
		return this.generateExpression(statement)
	default:
		panic(fmt.Sprintf (
			"BUG: generator doesnt know about statement %v, ty: %T",
			statement, statement))
	}
}

func (this *generator) generateStatementLoc (statement entity.Statement) (llvm.Value, error) {
	switch statement := statement.(type) {
	case entity.Expression:
		return this.generateExpressionLoc(statement)
	default:
		panic(fmt.Sprintf (
			"BUG: generator doesnt know about statement %v, ty: %T",
			statement, statement))
	}
}
