package generator

import "fmt"
import "git.tebibyte.media/sashakoshka/fspl/llvm"
import "git.tebibyte.media/sashakoshka/fspl/entity"
import "git.tebibyte.media/sashakoshka/fspl/analyzer"

func (this *generator) generateVariableLoc (variable *entity.Variable) (llvm.Value, error) {
	return this.blockManager.variable(variable.Declaration)
}

func (this *generator) generateDeclarationLoc (declaration *entity.Declaration) (llvm.Value, error) {
	return this.blockManager.addDeclaration(declaration, nil)
}

func (this *generator) generateSliceLoc (slice *entity.Slice) (llvm.Value, error) {
	var err error
	var start, end, dataAddress llvm.Value
	var elementType llvm.Type
	var sizeType *llvm.TypeInt; {
		ty, err := this.typedef("Index")
		if err != nil { return nil, err }
		sizeType = ty.(*llvm.TypeInt)
	}

	if slice.Start != nil {
		start, err = this.generateExpression(slice.Start)
		if err != nil { return nil, err }
	}
	if slice.End != nil {
		end, err = this.generateExpression(slice.End)
		if err != nil { return nil, err }
	}

	sourceType := analyzer.ReduceToBase(slice.Slice.Type())
	switch sourceType := sourceType.(type) {
	case *entity.TypeSlice:
		source, err := this.generateExpressionLoc(slice.Slice)
		if err != nil { return nil, err }
		irSourceType, err := this.generateType(sourceType)
		if err != nil { return nil, err }
		elementType, err = this.generateType(sourceType.Element)
		if err != nil { return nil, err }
		dataAddressFieldAddress := this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, 0))
		dataAddress = this.blockManager.NewLoad (
			new(llvm.TypePointer),
			dataAddressFieldAddress)
		lengthAddress := this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, 1))
			
		if end == nil {
			end = this.blockManager.NewLoad(sizeType, lengthAddress)
		}
		
	case *entity.TypeArray:
		elementType, err = this.generateType(sourceType.Element)
		if err != nil { return nil, err }
		dataAddress, err = this.generateExpressionLoc(slice.Slice)
		if err != nil { return nil, err }
		if end == nil {
			end = llvm.NewConstInt(sizeType, int64(sourceType.Length))
		}
	
	default:
		panic(fmt.Sprint (
			"BUG: generator can't slice expression ",
			slice.Slice))
	}

	// figure out starting position and length
	var length llvm.Value
	if start == nil {
		length = end
	} else {
		length = this.blockManager.NewSub(end, start)
		dataAddress = this.blockManager.NewGetElementPtr (
			elementType,
			dataAddress,
			start)
	}

	// create new slice descriptor
	destinationType, err := this.generateType(slice.Type())
	if err != nil { return nil, err }
	newSlice := this.blockManager.newAllocaFront(destinationType)
	dataAddressFieldAddress := this.blockManager.NewGetElementPtr (
		destinationType,
		newSlice,
		llvm.NewConstInt(llvm.I32, 0),
		llvm.NewConstInt(llvm.I32, 0))
	lengthFieldAddress := this.blockManager.NewGetElementPtr (
		destinationType,
		newSlice,
		llvm.NewConstInt(llvm.I32, 0),
		llvm.NewConstInt(llvm.I32, 1))
	this.blockManager.NewStore(dataAddress, dataAddressFieldAddress)
	this.blockManager.NewStore(length,      lengthFieldAddress)

	return newSlice, nil
}

func (this *generator) generateSubscriptLoc (subscript *entity.Subscript) (llvm.Value, error) {
	source, err := this.generateExpressionLoc(subscript.Slice)
	if err != nil { return nil, err }
	offset, err := this.generateExpression(subscript.Offset)
	if err != nil { return nil, err }

	var elementType entity.Type
	var dataAddress llvm.Value

	sourceType := analyzer.ReduceToBase(subscript.Slice.Type())
	switch sourceType := sourceType.(type) {
	case *entity.TypeSlice:
		irSourceType, err := this.generateType(sourceType)
		if err != nil { return nil, err }
		
		// get slice element
		elementType = sourceType.Element
		dataAddressFieldAddress := this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, 0))
		dataAddress = this.blockManager.NewLoad (
			new(llvm.TypePointer),
			dataAddressFieldAddress)
		
	case *entity.TypeArray:
		elementType = sourceType.Element
		dataAddress = source
	
	default:
		panic(fmt.Sprint (
			"BUG: generator can't subscript expression ",
			subscript.Slice))
	}

	irElementType, err := this.generateType(elementType)
	if err != nil { return nil, err }
	return this.blockManager.NewGetElementPtr (
		irElementType,
		dataAddress,
		offset), nil
}

func (this *generator) generateDereferenceLoc (dereference *entity.Dereference) (llvm.Value, error) {
	return this.generateExpression(dereference.Pointer)
}

func (this *generator) generateMemberAccessLoc (access *entity.MemberAccess) (llvm.Value, error) {
	source, err := this.generateExpressionLoc(access.Source)
	if err != nil { return nil, err }

	switch sourceTypeAny := analyzer.ReduceToBase(access.Source.Type()).(type) {
	case *entity.TypeStruct:
		sourceType := sourceTypeAny
		irSourceType, err := this.generateType(sourceType)
		if err != nil { return nil, err }
		offset := this.getStructMemberIndex(sourceType, access.Member)
		return this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, int64(offset))), nil
		
	case *entity.TypePointer:
		sourceType := analyzer.ReduceToBase(sourceTypeAny.Referenced).(*entity.TypeStruct)
		irSourceType, err := this.generateType(sourceType)
		if err != nil { return nil, err }
		offset := this.getStructMemberIndex(sourceType, access.Member)
		source = this.blockManager.NewLoad(new(llvm.TypePointer), source)
		return this.blockManager.NewGetElementPtr (
			irSourceType,
			source,
			llvm.NewConstInt(llvm.I32, 0),
			llvm.NewConstInt(llvm.I32, int64(offset))), nil
		
	default:
		panic(fmt.Sprint (
			"BUG: generator can't access members in expression ",
			access.Source))
	}
}

func (this *generator) generateBlockLoc (block *entity.Block) (llvm.Value, error) {
	var last llvm.Value
	var err error
	for index, step := range block.Steps {
		// prefer generating a normal statement to avoid allocations
		if index == len(block.Steps) - 1 {
			last, err = this.generateStatementLoc(step)
		} else {
			last, err = this.generateStatement(step)
		}
		if err != nil { return nil, err }
	}
	return last, nil
}

